# Copyright (C) 2019 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from majormode.perseus.client.api import PerseusConnection
from majormode.perseus.client.service.base_service import BaseService
from majormode.perseus.constant.account import AccountType
from majormode.perseus.model import obj
from majormode.perseus.model.contact import Contact
from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.model.obj import Object

import getpass
import json
import os


class AccountService(BaseService):
    class AuthenticationFailureException(BaseService.BaseServiceException):
        """
        Signal that the email address or password that has been provided for
        authenticating a user account is incorrect.
        """
        pass

    class ContactAlreadyUsedException(BaseService.BaseServiceException):
        """
        Signal that one or more contact information such as, for instance, an
        email address, provided to sign-up a user account, is already used by
        another user account of the platform.
        """
        pass

    class UnsupportedOAuthServiceProvider(BaseService.BaseServiceException):
        """
        Signal that the specified OAuth Service Provider is not yet supported
        by the platform.
        """
        pass

    BaseService._declare_custom_exceptions({
        'AuthenticationFailureException': AuthenticationFailureException,
        'ContactAlreadyUsedException': ContactAlreadyUsedException,
        'UnsupportedOAuthServiceProvider': UnsupportedOAuthServiceProvider
    })

    def __init__(self, session):
        super().__init__(session)

    def is_contact_available(self, contact):
        """
        Indicate whether the specified contact information is available.

        In the case this contact information would have been already
        registered by a user, indicate whether this contact information has
        been also verified.


        :param contact: An instance `Contact`


        :return: A tuple containing the following values:

            * `is_available`: `True` if the specified contact is not registered by
              any existing account; `False` otherwise.

            * `is_verified`: `True` if the specified contact has been registered
              and verified; `False` otherwise.
        """
        payload = Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/account/contact/(name)/availability',
                url_bits={
                    'name': str(contact.name)
                },
                arguments={
                    'value': contact.value
                },
                authentication_required=False,
                signature_required=True))

        return payload.is_available, payload.is_verified

    def set_fullname(self, fullname):
        """
        Update the complete full name of a user.


        :param fullname: Full name by which the user is known, as given by the
            user himself, i.e., untrusted information.


        :return: An instance containing the following attributes:

            * ``account_id`` (required): Identification of the user's account.

            * ``fullname`` (required): The new full name of the user.

            * ``update_time`` (required): Time of the most recent modification of
              the attributes of the user's account.


        @raise DeletedObjectException: If the user's account has been deleted.

        @raise DisabledObjectException: If the user's account has been
            disabled.

        @raise Undefined ObjectException: If the specified identification
            doesn't refer to a user account registered to the platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.PUT,
                path='/account/fullname',
                message_body={
                    'fullname': fullname
                },
                authentication_required=True,
                signature_required=True))

    def upload_avatar(self, file):
        """
        Upload a new picture of this user account, also known as the avatar,
        which is the graphical representation of the user.


        :param file: A file-like object to upload to the platform.


        :return: An instance containing the following attributes:

            * ``file_name`` (required): The original local file name as the
            ``filename`` parameter of the ``Content-Disposition`` header.

            * ``picture_id`` (required): Identification of the new avatar of the
                user's account as registered to the platform.

            * ``picture_url`` (required): Uniform Resource Locator (URL) that
              specifies the location of the user's avatar.  The client application
              can use this URL and append the query parameter ``size`` to specify
              a given pixel resolution of the photo, such as ``thumbnail``,
              ``small``, ``medium``, or ``large``.

            * ``update_time`` (required): Time of the most recent modification of
              the attributes of the user's account.  This information should be
              stored by the client application to manage its cache of user
              accounts.


        @raise DeletedObjectException: If the user's account has been deleted.

        @raise DisabledObjectException: If the user's account has been
            disabled.

        @raise InvalidOperation: If the format of the uploaded image is not
            supported.

        @raise Undefined ObjectException: If the specified identification
            doesn't refer to a user account registered to the platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.PUT,
                path='/account/avatar',
                files=[file],
                authentication_required=True,
                signature_required=True))
































    def add_contacts(self, contacts):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/account/contacts',
                message_body=contacts,
                authentication_required=False,
                signature_required=True))

    def __sign_in(self, email_address, password):
        """
        Sign-in the user against the platform providing his email address and
        his password.


        :param email_address: email address of the user.

        :param password: password of the user.


        :return: a session instance containing the following members:

            * ``account_id`` (required): identification of the account of the
              user.

            * ``contacts`` (required): a list of contact information that can be
              used as primary contact information for this user account.  A
              contact information corresponds to the following tuples::

                   ``(name, value, is_verified)``

                where:

                * ``name``: an item of the enumeration ``ContactName``.

                * ``value``: value of the property representing by a string.

                * ``is_verified``: indicate whether this contact information has been
                  verified.

            * ``expiration_time``: time when the login session will expire.

            * ``fullname``: complete full name of the user as given by the user
              himself, i.e., untrusted information, or as determined from his
              email address as for a ghost account.

            * ``session_id``: identification of the login session of the user.
        """
        payload = Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/account/session',
                message_body={
                    'email_address': email_address,
                    'password': password
                },
                authentication_required=False,
                signature_required=True))

        self.set_session(self.session.connection.build_authenticated_session(payload))

        return self.session

    def get_account(self, account_id, include_pending=False):
        accounts = Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/account',
                arguments={
                    'ids': account_id,
                    'include_pending': include_pending
                },
                authentication_required=False,
                signature_required=True))

        if len(accounts) == 0:
            raise self.UndefinedObjectException()

        return accounts[0]

    def get_contact_verification_request(self, request_id):
        """

        :param request_id:


        :return:
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/account/contact/verification/(request_id)',
                url_bits={
                    'request_id': request_id
                },
                authentication_required=False,
                signature_required=True))

    def get_password_reset_request(
            self, request_id,
            check_access=False,
            check_app=False,
            check_status=False):
        """
        Return extended information about the specified password reset that a
        user has requested for his account.


        :param request_id: identification of the password request.

        :param check_access: indicate whether the function must check if the
            user on behalf of whom the function is called is either the user
            who added this photo.

        :param check_app: indicate whether the function must check if the
            client application on behalf of which the function is called is
            the same than the client application that requested the password
            reset of the user's account.

        :param check_status: indicate whether the function must check the
            current status of this passwor reset request and raise an
            exception if this status is not of enabled.


        :return: an instance containing the following members:

            * ``account_id`` (required): identification of the user's account
              which the user has requested to reset his password.

            * ``app_id`` (required): identification of the client application that
              submitted on behalf of the user the request to reset the password
              of his account.

            * ``attempt_count`` (required): number of times the platform sent an
              email to the user with an embedded link to let this user reset his
              password.

            * ``creation_time`` (required): time when the user requested to reset
              the password of his account.

            * ``request_count`` (required): number of times the user requested to
              reset his password before he finally changed it.

            * ``request_id`` (required): identification of this password reset
              request.

            * ``update_time`` (required): the most recent time when the platform
              sent an email to the user with an embedded link to let this user
              reset his password.
        """
        payload = Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/account/password/reset/request/(request_id)',
                url_bits={
                    'request_id': request_id
                },
                arguments={
                    'check_access': check_access,
                    'check_app': check_app,
                    'check_status': check_status
                },
                authentication_required=False,
                signature_required=True))

        return payload

    def get_sns_data_deletion(self, request_id):
        """
        Return information about a request sent to the platform to delete the
        data of a user that an application has collected from the given Social
        Networking Service about this user.


        :param request_id: identification of the data deletion request as
            registered to the platform.


        :return: an instance containing the following attributes:

            * ``creation_time`` (required): time when the request to delete the data
              of a user has been initiated.

            * ``object_status`` (required): current status of this data deletion
              request.

            * ``request_id`` (required): the identification of the data deletion
              request as registered to the platform.

            * ``sns_app_id`` (required): identification of the application as
              registered to the 3rd party Social Networking Service.

            * ``sns_name`` (required): code name of the 3rd party Social Networking
              Service.

            * ``sns_user_id`` (required): identification of the user as registered
              to the 3rd party Social Networking Service.

            * ``update_time`` (required): time of the most recent update of the
              access token of this user.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/account/sns/deletion/request/(request_id)',
                url_bits={
                    'request_id': request_id
                },
                authentication_required=False,
                signature_required=True))



    def is_contact_verification_request(self, name, request_id):
        """
        Check whether the given identification corresponds to a verification
        request of the specified type of contact.


        :param name: an instance `Contact.ContactName` of the contact that is
            requested to be verified.

        :param request_id: identification of the contact verification request.


        :return: `True` if the specified identification corresponds to a
            contact verification request registered to the platform;
            ``False`` otherwise.
        """
        payload = Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/account/(name)/verification/(request_id)',
                url_bits={
                    'name': name,
                    'request_id': request_id
                },
                authentication_required=False,
                signature_required=True))

        return payload.data

    def request_contact_verification(self, contact, locale=DEFAULT_LOCALE):
        """
        Request the platform to initiate the process to verify the specified
        contact information.

        The function generates a *nonce* ("number used once"), which is a
        pseudo-random number issued to identify the verification of the
        contact information.  To confirm the contact information of the user,
        the function ``confirm_contact`` will have to be called with this
        *nonce*.


        :param contact: an instance `Contact`.

        :param locale: a ``Locale`` instance referencing the preferred
            language of the user.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/account/(name)/verification',
                url_bits={
                    'name': contact.name
                },
                message_body={
                    'locale': locale,
                    'value': contact.value
                },
                authentication_required=False,
                signature_required=True))

    def request_password_reset(self, email_address):
        """
        Request the platform to initiate the process to help the user in
        resetting his password that he has forgotten.

        The function generates a *nonce* ("number used once"), which is a
        pseudo-random number issued to identify the request of the user.  This
        nonce will be sent to the user to his email address, more likely in a
        HTML link that will redirect the end user to a web page responsible
        for allowing the user to reset his password (cf. function
        ``change_forgotten_password`` that must be passed this *nonce*).

        Note: if the user sends consecutively two requests to reset his
        password within the minimal allowed duration of time, the function
        ignores the new request and returns the identification of the previous
        request.


        :param email_address: email address of the account of the user who is
            requesting to reset is forgotten password.


        @raise DeletedObjectException: if the user account has been deleted.

        @raise DisabledObjectException: if the user account has been disabled.

        @raise InvalidOperationException: if a password reset has already been
            requested recently for this email address.

        @raise UndefinedObjectException: if the specified email address
            doesn't refer to any user account registered against the platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/account/password/reset/request',
                message_body={
                    'email_address': email_address
                },
                authentication_required=False,
                signature_required=True))

    def request_sns_data_deletion(self, sns_name, sns_app_id, sns_user_id):
        """
        Request the platform to delete the data of the specified user that an
        application has collected from the given Social Networking Service
        about this user.


        :param sns_name: code name of the 3rd party Social Networking Service.

        :param sns_app_id: identification of the application as registered to
            the 3rd party Social Networking Service.

        :param sns_user_id: identification of the user as registered to the
            3rd party Social Networking Service.


        :return: an instance containing the following attributes:

            * ``creation_time`` (required): time when this request has been
              registered to the platform.

            * ``request_id`` (required): identification of the data deletion
              request as registered to the platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/account/sns/deletion/request',
                message_body={
                    'sns_app_id': sns_app_id,
                    'sns_name': sns_name,
                    'sns_user_id': sns_user_id
                },
                authentication_required=False,
                signature_required=True))

    def reset_password(self, request_id, new_password):
        """
        Change the password of the account of a user who forgot his password
        and who requested the platform to reset, as this user cannot login
        into the platform anymore.


        :param request_id: identification of the request of the user to reset
            his forgotten that the function ``request_password_reset``
            generated.

        :param new_password: new password of this user's account.


        @raise InvalidArgumentException: if the new password doesn't conform
            to the rules of password definition.

        @raise UndefinedObjectException: if the specified identification
            doesn't refer to a request from a user to reset his forgotten
            password.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/account/password/reset',
                message_body={
                    'new_password': new_password,
                    'request_id': request_id
                },
                authentication_required=False,
            signature_required=True))

    def sign_out(self):
        """
        Sign out the user from his login session.


        @raise IllegalAccessException: if the specified login session doesn't
            belong to the specified user.

        @raise UndefinedObjectException: if the specified identification
            doesn't refer to any user login session registered against the
            platform.
        """
        payload = Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.DELETE,
                path='/account/session',
                authentication_required=True,
                signature_required=True))

        self.flush_session()

        return payload

    def sign_up_with_oauth(
            self, provider_name, user_id,
            oauth_consumer_key, oauth_consumer_secret, access_token,
            auto_sign_in=False):
        """
        Sign-up the user against the platform providing the OAuth access token
        used by the Consumer to gain access to to the Protected Resources on
        behalf of the User


        :param provider_name: code name of the Service Provider, such as, for
            instance, ``facebook``, ``twitter``, etc.

        :param user_id: identification of the 3rd party user account on the
            Service Provider.

        :param oauth_consumer_key: a value used by the Consumer to identify
            itself to the Service Provider.

        :param oauth_consumer_secret: a secret used by the Consumer to
            establish ownership of the Consumer Key.

        :param access_token: a value used by the Consumer to gain access to
            the Protected Resources on behalf of the User, instead of using
            the User’s Service Provider credentials.

        :param auto_sign_in: indicate whether the platform is requested to
            sign-in this user once the sign-up procedure completes
            successfully.


        @raise IllegalAccessException: if the identifier of the user that the
            specified OAuth token impersonates is not the identifier that is
            passed to the function.
        """
        session_payload = Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/account',
                arguments={
                    'auto_sign_in': auto_sign_in
                },
                message_body={
                    'access_token': access_token,
                    'oauth_consumer_key': oauth_consumer_key,
                    'oauth_consumer_secret': oauth_consumer_secret,
                    'provider_name': provider_name,
                    'user_id': user_id
                },
                authentication_required=False,
                signature_required=True))

        if auto_sign_in:
            self.set_session(self.session.connection.build_authenticated_session(session_payload))

        return session_payload

    @staticmethod
    def terminal_login(
            connection=None,
            reset_session_info=False,
            store_session_info=True,
            session_file_path_name=None):
        """
        Connect and log into a server platform as specified with connection
        properties, requesting the user to prove his identity to the server
        platform by providing his credentials.

        If a session has been already stored in the specified file, the
        function retrieves the connection and session properties from this
        file, update the connection instance ``connection`` passed to
        this function, and return a session instance built with these stored
        session properties.


        :param connection: an instance of ``PerseusConnection``
            containing the connection properties to the server platform.

        :param reset_session_info: indicate whether to reset session
            information that would have been previously stored on the local
            disk in the file specified by its path and name.

        :param store_session_info: indicate whether to store the session on
            the local disk with the specified path and file name.

        :param session_file_path_name: the path and the name of the file where
            the session information, and possibly the connection information,
            will be saved in.


        :return: An authenticated session of the user logged to the RESTful
            API server.
        """
        session = None

        # If the hostname and the port number of the Perseus RESTful API server
        # have not been defined, request the user to enter them.
        hostname = connection.hostname if connection and connection.hostname \
            else input("Enter the RESTful API server's hostname: ")

        port_number = connection.port_number if connection and connection.port_number \
            else PerseusConnection.DEFAULT_PORT_NUMBER

        # Check whether a login session of the user is already stored for this
        # RESTful API server.
        if session_file_path_name is None:
            session_file_path_name = '.%s.session' % hostname

        if os.path.exists(session_file_path_name):
            if reset_session_info:
                os.remove(session_file_path_name)
            else:
                try:
                    with open(session_file_path_name) as fd:
                        payload = json.loads(fd.read())
                        connection = PerseusConnection.from_json(hostname, payload, port_number=port_number)
                        return connection.build_session(payload)
                except ValueError:
                    pass

        # If the consumer key and secret have not been passed to the function,
        # request the user to enter them.
        consumer_key = connection.consumer_key if connection and connection.consumer_key \
            else input('Enter your Consumer Key: ')

        consumer_secret = connection.consumer_secret if connection and connection.consumer_secret \
            else getpass.getpass('Enter your Consumer Secret: ')

        if connection is None:
            connection = PerseusConnection(
                hostname,
                consumer_key,
                consumer_secret,
                port_number=port_number)

        # If the user is not already logged, request the user to provide his
        # credentials.
        if session is None:
            email_address = input("Enter your email address: ")
            password = getpass.getpass()
            session = AccountService(connection).__sign_in(email_address, password)

            if store_session_info:
                payload = Object()
                payload.merge(connection)
                payload.merge(session)

                with os.fdopen(os.open(session_file_path_name, os.O_WRONLY | os.O_CREAT, 0o600), 'w') as handle:
                    handle.write(json.dumps(obj.stringify(payload)))

        return session



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from majormode.perseus.client.service.team.team_service import TeamService
from majormode.perseus.client.unittest.base_unittest import BaseApiKeyTestCases

import unittest
import uuid

class TeamServiceTestCase(BaseApiKeyTestCase):
    def test_add_team(self):
        TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)

    def test_add_duplicated_team(self):
        name = 'test_%s' % uuid.uuid1().hex
        TeamService().add_team(self.unittest_app_id, self.unittest_account_id, name)
        self.assertRaises(TeamService.NameAlreadyUsedException,
            TeamService().add_team,
            self.unittest_app_id, self.unittest_account_id, name)

    def test_get_account_teams(self):
        TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)
        TeamService().get_account_teams(self.unittest_app_id, self.unittest_account_id)

    def test_add_members(self):
        team = TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)
        account = AccountServiceTestCase.generate_account(self.unittest_app_id)
        TeamService().add_members(self.unittest_app_id, self.unittest_account_id,
            team.team_id, [ account.email_address ])

    def test_add_duplicated_master_members(self):
        team = TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)

        account = AccountService().get_account(self.unittest_account_id)
        TeamService().add_members(self.unittest_app_id, self.unittest_account_id,
            team.team_id, [ account.email_address ])

    def test_add_duplicated_members(self):
        team = TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)

        account = AccountServiceTestCase.generate_account(self.unittest_app_id)
        TeamService().add_members(self.unittest_app_id, self.unittest_account_id,
            team.team_id, [ account.email_address ])
        TeamService().add_members(self.unittest_app_id, self.unittest_account_id,
            team.team_id, [ account.email_address ])

    def test_accept_member_request(self):
        team = TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)
        account = AccountServiceTestCase.generate_account(self.unittest_app_id)
        TeamService().add_members(self.unittest_app_id, self.unittest_account_id,
            team.team_id, [ account.email_address ])
        request_secured_key = TeamService()._get_request_secured_key(team.team_id, account.account_id)
        TeamService().accept_member_request(self.unittest_app_id, request_secured_key)

    def test_decline_member_request(self):
        team = TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)
        account = AccountServiceTestCase.generate_account(self.unittest_app_id)
        TeamService().add_members(self.unittest_app_id, self.unittest_account_id,
            team.team_id, [ account.email_address ])
        request_secured_key = TeamService()._get_request_secured_key(team.team_id, account.account_id)
        TeamService().decline_member_request(self.unittest_app_id, request_secured_key)

    def test_revoke_member(self):
        team = TeamService().add_team(self.unittest_app_id, self.unittest_account_id,
            'test_%s' % uuid.uuid1().hex)
        account = AccountServiceTestCase.generate_account(self.unittest_app_id)
        TeamService().add_members(self.unittest_app_id, self.unittest_account_id,
            team.team_id, [ account.email_address ])
        request_secured_key = TeamService()._get_request_secured_key(team.team_id, account.account_id)
        TeamService().accept_member_request(self.unittest_app_id, request_secured_key, account.account_id)

        TeamService().revoke_member(self.unittest_app_id, self.unittest_account_id,
            team.team_id, account.account_id)
        self.assertRaises(TeamService.InvalidOperationException, TeamService().revoke_member,
            self.unittest_app_id, self.unittest_account_id,
            team.team_id, account.account_id)

def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(TeamServiceTestCase))
    return test_suite

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from majormode.perseus.client.service.base_service import BaseService
from majormode.perseus.model.obj import Object


class TeamService(BaseService):
    BaseService._declare_custom_exceptions({
    })

    def __init__(self, connection):
        super(TeamService, self).__init__(connection)

    def accept_invite(self, invite_code):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/team/invite/(invite_code)/acceptance',
                url_bits={'invite_code': invite_code},
                authentication_required=False,
                signature_required=True))


    def add_team(self, name, description=None,
            invite_url=None, invite_email=None):
        """
        Add a new team on behalf of a user who becomes then the de facto
        master administrator for this team.

        @param app_id: identification of the client application such as a Web,
                a desktop, or a mobile application, that accesses the service.
        @param account_id: identification of the account of a user who add a
               team.
        @param name: name of the team.  This name MUST be unique among all the
               teams that have been registered so far against the platform.
               Case is not sensitive.
        @param description: a short textual description of the team, if any
               provided.
        @param invite_url: Uniform Resource Locator (URL) that is provided as
               a link in the email the platform sends to a user who is invited
               to join the team.  When the user clicks on the link embedded in
               the email, the email reader application issues a HTTP GET
               request to this URL.
        @param invite_email: template of the letter to be sent by email to a
               user who is invited to join a team.  If no specific template is
               specified for this team, the platform provides a default
               template.

        @raise DeletedObjectException: if the user account has been deleted.
        @raise DisabledObjectException: if the user account has been disabled.
        @raise NameAlreadyUsedException: if the name of the team passed to
               this function is already registered for an other team.
        @raise UndefinedObjectException: if the specified identification
               doesn't refer to a user account registered against the
               platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/team',
                message_body={
                    'name': name,
                    'description': description,
                    'invite_url': invite_url,
                    'invite_email': invite_email
                },
                authentication_required=True,
                signature_required=True))


    def cancel_invite(self, invite_id):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.DELETE,
                path='/team/invite/(invite_id)',
                url_bits={'invite_id': invite_id},
                authentication_required=True,
                signature_required=True))


    def decline_invite(self, invite_code):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/team/invite/(invite_code)/declination',
                url_bits={'invite_code': invite_code},
                authentication_required=False,
                signature_required=True))


    def get_member(self, team_id, account_id):
        """
        Indicate whether the authenticated user belongs to a given team.

        @param team_id: identification of a team.

        @param account_id: identification of the account of the user to check
            his membership to the specified team.

        @return: an instance containing the following members:

        @raise DeletedObjectException: if the team has been deleted.

        @raise DisabledObjectException: if the team has been disabled.

        @raise IllegalAccessException: if the user on behalf of the request is
            is not a member of this team.

        @raise UndefinedObjectException: if the specified identification
            doesn't refer to a team registered against the platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/team/(team_id)/member/(account_id)',
                url_bits={
                    'account_id': account_id,
                    'team_id': team_id },
                authentication_required=True,
                signature_required=True))


    def get_members(self, team_id,
            limit=BaseService.DEFAULT_RESULT_SET_SIZE, offset=0):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/team/(team_id)/member',
                url_bits={'team_id': team_id },
                arguments={
                    'limit': limit,
                    'offset': offset,
                },
                authentication_required=True,
                signature_required=True))


    def get_account_teams(self, limit=BaseService.DEFAULT_RESULT_SET_SIZE, offset=0):
        """
        Return a list of teams the authenticated account is member of.

        @param offset: require to skip that many teams before beginning to
               return teams.
        @param limit: constrain the number of teams that are returned to the
               specified number.

        @return: an array of instance containing the following members:
                 * ``team_id``: identification of an team the
                   specified user is member of.
                 * ``name``: the title of the team.
                 * ``description``: a short textual description of the
                   team, if any provided.
                 * ``creation_time``: date and time when the team has
                   been registered.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/team',
                arguments={
                    'limit': limit,
                    'offset': offset,
                },
                authentication_required=True,
                signature_required=True))


    def invite_users(self, team_id, account_ids):
        """
        Invite a list of users to join the specified team.  The platform will
        eventually send them an invitation, either by an email or an in-app
        notification.  Their membership is pending until they accept or
        decline the invitation.

        The function filters out email addresses of users who are already
        members of the team and email addresses of users who have been already
        invited but who have not yet accepted.

        @note: only an administrator of the team is allowed to add members to
               a team.

        @note: the function silently creates ghost accounts for users who have
               not an email address registered against the platform yet.  The
               locale that is used to reference the preferred language of such
               users is the same than for the administrator of the team.


        @param team_id: identification of the team to add the specified user
               as a new member.
        @param account_ids: a list of valid email addresses of the users
               to add as new members to the specified team.

        @return: the list of email addresses of new members who have been
                 added to the team; users who were already members of the team
                 are filtered out.

        @raise DeletedObjectException: if the team has been deleted.
        @raise DisabledObjectException: if the team has been disabled.
        @raise IllegalAccessException: if the user on behalf of the new member
               is added to the team is not the master administrator or one of
               the regular administrators of this team.
        @raise InvalidArgumentException: if some email addresses are of a
               wrong format, i.e., not compliant with RFC 2822.
        @raise InvalidOperationException: if no URL callback for accepting
               member request has been defined for this team.
        @raise UndefinedObjectException: if the specified identification
               doesn't refer to a team registered against the platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/team/(team_id)/invite',
                url_bits={'team_id': team_id},
                message_body=account_ids,
                authentication_required=True,
                signature_required=True))


    def is_administrator(self, team_id):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.HEAD,
                path='/team/(team_id)/administrator',
                url_bits={
                    'team_id': team_id
                },
                authentication_required=False,
                signature_required=True))


    def revoke_member(self, team_id, account_id):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.DELETE,
                path='/team/(team_id)/member/(account_id)',
                url_bits={
                    'account_id': account_id,
                    'team_id': team_id
                },
                authentication_required=True,
                signature_required=True))


    def search_teams(self, keywords, limit=BaseService.DEFAULT_RESULT_SET_SIZE, offset=0):
        """
        Return a list of teams which names match, even partially, the
        specified keywords.

        @param keywords: a list of keywords.

        @param limit: constrain the number of teams that are returned to the
            specified number.

        @param offset: require to skip that many teams before beginning to
            return teams.

        @return: an instance containing the following members:

            * ``team_id``: identification of the team.

            * ``name``: the name of the team.

            * ``description``: a short textual description of the team, if any
              provided.

            * ``account_id``: identification of the account of the agent for
              this team.

            * ``picture_id``: identification of the picture that represents
              the team, if any picture defined for this team.

            * ``picture_url``: Uniform Resource Locator (URL) that specifies
              the location of the picture representing the team, if any
              defined.  The client application can use this URL and append the
              query parameter ``size`` to specify a given pixel resolution of
              the team's picture:

               * ``thumbnail``

               * ``small``

               * ``medium``

               * ``large``

            * ``creation_time``: time when the team has been registered.

            * ``update_time``: most recent time when some information, such as
              the name or the description of the team, has been modified.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/team',
                arguments={
                    'keywords': keywords,
                    'limit': limit,
                    'offset': offset,
                },
                authentication_required=False,
                signature_required=True))


    def submit_join_request(self, team_id):
        """
        Submit a request on behalf of a user to join a team.  This join
        request has to be approved by one of the administrators of this team.

        @param team_id: identification of the team the user requests to join.

        @raise DeletedObjectException: if the team has been deleted.

        @raise DisabledObjectException: if the team has been disabled.

        @raise UndefinedObjectException: if the specified identification
            doesn't refer to a team registered against the platform.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.POST,
                path='/team/(team_id)/join',
                url_bits={'team_id': team_id},
                authentication_required=True,
                signature_required=True))



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import random
import settings
import unittest


class BaseTestCase(unittest.TestCase):
    def setUp(self):
#        logging.basicConfig(stream=sys.stdout)
#        logging.getLogger('UnitTest').setLevel(logging.DEBUG)
#        self.logger = logging.getLogger('UnitTest')

        random.seed()

        self.unittest_account_id = settings.PLATFORM_UNITTEST_ACCOUNT_ID
        self.unittest_app_id = settings.PLATFORM_UNITTEST_APP_ID

    def tearDown(self):
        pass

class BaseApiKeyTestCase(BaseTestCase):
    def setUp(self):
        super(BaseApiKeyTestCase, self).setUp()
        self.api_key = settings.PLATFORM_UNITTEST_API_KEY
        self.secret_key = settings.PLATFORM_UNITTEST_SECRET_KEY
        self.account_id = settings.PLATFORM_UNITTEST_ACCOUNT_ID

def suite():
    return unittest.TestSuite()



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import datetime
import hashlib
import hmac
import json
import logging
import re
import requests
import socket
import urllib.parse
import urllib.request
import uuid
import traceback

from majormode.perseus.constant.http import HttpMethod
from majormode.perseus.constant.http import HttpScheme
from majormode.perseus.model import obj
from majormode.perseus.model.obj import Object
from majormode.perseus.utils import cast


class PerseusConnection:
    """
    Connection properties to access a RESTful API server built on the
    Perseus server framework.
    """
    DEFAULT_HTTP_PORT_NUMBER = 80
    DEFAULT_HTTPS_PORT_NUMBER = 443
    DEFAULT_PORT_NUMBER = DEFAULT_HTTPS_PORT_NUMBER

    def __build_anonymous_session(self):
        """
        Build an anonymous user session to the RESTful API server identified
        by this connection.


        :return: An object `PerseusAnonymousSession` connected to the RESTful
            API server identified by this connection.
        """
        return PerseusAnonymousSession(self)

    def __build_authenticated_session(self, payload, strict=True):
        """
        Build an authenticated user session to the RESTful API server
        identified by this connection.

        The function retrieves the session properties from the specified
        payload.


        :param payload: the payload of the user session provided as:

            * A dictionary

            * A string representation of a JSON object

            * An object `Object`


        :return: An objet `PerseusAuthenticatedSession`.
        """
        try:
            return PerseusAuthenticatedSession(self, payload) if isinstance(payload, Object) else \
                PerseusAuthenticatedSession.from_payload(
                    self,
                    cast.string_to_json(payload) if isinstance(payload, str) else payload)
        except KeyError as exception:
            if strict:
                raise exception
            else:
                logging.debug(traceback.format_exc())
                return self.__build_anonymous_session()

    def __init__(
            self,
            hostname,
            consumer_key,
            consumer_secret,
            api_version=None,
            port_number=None,
            debug=False,
            timeout=None,
            radioactive_tracer=False):
        """
        Build an object `PerseusConnection` defining the properties to connect
        to a RESTful API server built on the Perseus server framework.


        :param hostname: Fully qualified domain name of the RESTful API
            server.

        :param consumer_key: A value used by the consumer to identify itself
            to the service provider.

        :param consumer_secret: Secret used by the consumer to establish
            ownership of the consumer key.

        :param api_version: An object `Version` that indicates the version
            of the platform RESTful API that is integrated.

        :param port_number: Port number, added to the RESTful API server's IP
            address, to complete the destination address for the communication
            session.  That is, data packets are routed across the network to
            the specific destination IP address, and then, upon reaching the
            destination computer, are further routed to the specific process
            bound to the destination port_number number.

        :param debug: Enable or disable debug mode.  If this is `True`, the
            cURL representation of a request sent to the RESTful API server is
            written to the standard input.

        :param timeout: Enforce the request to be processed within this
            predetermined elapsed time, expressed in seconds, rather than
            waiting indefinitely.  The request will be aborted after the
            timeout period has elapsed.  This is based on the assumption that
            further waiting is useless, and some other action is necessary.

        :param radioactive_tracer: Indicate whether the HTTP request needs
            to integrate a "radioactive tracer", i.e., a nonce ("number used
            once"), a pseudo-random number, used to explore the mechanism of
            subsequent procedure calls by tracing the path that the HTTP
            request follows.  The HTTP request is added an additional header
            `X-Radioactive-Tracer`.
        """
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.api_version = api_version
        self.hostname = hostname
        self.port_number = port_number or self.DEFAULT_PORT_NUMBER
        self.scheme = HttpScheme.https if self.port_number == self.DEFAULT_HTTPS_PORT_NUMBER else HttpScheme.http
        self.debug = debug
        self.timeout = timeout
        self.radioactive_tracer = radioactive_tracer

    def build_session(self, payload=None):
        """
        Build a user session to the RESTful API server identified by this
        connection.

        The user can be anonymous or authenticated.


        :param payload: A dictionary, or a string representation of a JSON
            expression, of the attributes of an existing session to this
            RESTful API server.


        :return: An object `PerseusAuthenticatedSession` if the payload of an
            existing authenticated user session was passed to this function,
            otherwise an object `PerseusAnonymousSession`.
        """
        return self.__build_anonymous_session() if not payload \
            else self.__build_authenticated_session(payload)

    @classmethod
    def from_json(cls, hostname, payload, port_number=None):
        """
        Build an object `PerseusConnection` from a JSON serialization of this
        object.


        :param hostname: Fully qualified domain name of the RESTful API
            server.

        :param payload: A JSON expression containing at least the following
            properties:

            * `consumer_key`: A value used by the consumer to identify itself to the
              service provider.

            * `consumer_secret`: Secret used by the consumer to establish ownership
              of the consumer key.

        :param port_number: Port number, added to the RESTful API server's IP
            address, to complete the destination address for the communication
            session.


        :return: An object `PerseusConnection`.
        """
        return PerseusConnection(
            hostname,
            payload['consumer_key'],
            payload['consumer_secret'],
            port_number=payload.get('port_number') or port_number or cls.DEFAULT_PORT_NUMBER)


class PerseusSession:
    """
    Session of user connected to a RESTful API server built on the Perseus
    server framework.
    """
    CRYPTOGRAPHIC_HASH_FUNCTION = hashlib.sha1

    class Request(urllib.request.Request):
        """
        Support of the explicit specification of a HTTP method.
        """
        def __init__(self, *args, **kwargs):
            http_method = kwargs.pop('http_method', None)
            self.__method = http_method and str(http_method)
            urllib.request.Request.__init__(self, *args, **kwargs)

        def get_method(self):
            return self.__method or super().get_method()

    class SocketTimeoutException(Exception):
        """
        Signal that a timeout occurs on a socket during blocking operations
        like the connection attempt of a HTTP request.
        """

    def __build_curl_command(self, http_method, path_query, message_body, headers):
        """
        Build the string representation of the cURL command-line that sends
        this request.


        :param http_method: An item of the enumeration `HttpMethod`.

        :param path_query: A string representation of the path and the query
            parameters of the URL endpoint.

        :param message_body: A string representation of the message body to
            pass along the HTTP request when the HTTP method `POST` or `PUT`
            is used.

        :param headers: A dictionary of headers to pass along the HTTP request.


        :return: A string representation of the cURL command-line that sends
            this request.
        """
        url_string = f'{self.connection.scheme}://{self.connection.hostname}:{self.connection.port_number}{path_query}'

        headers_string = ' '.join([
            f'-H "{key}: {value}"'
            for key, value in headers.items()
            if value is not None
        ])

        curl_string = f'curl -X {http_method} "{url_string}" {headers_string}'

        if http_method in (HttpMethod.POST, HttpMethod.PUT):
            curl_string += f" -d '{message_body or ''}'"

        return curl_string

    @classmethod
    def __build_path_query_string(cls, path, arguments):
        """
        Build the URL path with its query string.


        The query string is built with the specified list of arguments.  For
        each argument that would correspond to a list of items, the value
        of this argument is built with the concatenation of the string
        representation of each item separated with a comma.

        For instance, the query string built from the following arguments:

            {
              "foo": uuid.UUID('2981c498-694a-11e3-ba63-00224d868489'),
              "bar": [1, 2, 3, 5, 7, 11]
            }

        is:

            foo=2981c498-694a-11e3-ba63-00224d868489&bar=1,2,3,4,5,7,11


        :param path: The path to the URL endpoint.

        :param arguments: A dictionary corresponding to the query parameters
            to pass to the URL endpoint.


        :return:
        """
        if arguments is None:
            return path

        stringified_arguments = {
            key: cls.__stringify_query_parameter_value(value)
            for key, value in arguments.items()
            if value is not None
        }

        encoded_query_string = urllib.parse.urlencode(stringified_arguments)

        path_query_string = f'{path}?{encoded_query_string}'
        return path_query_string

    def __calculate_signature(self, path_query, message_body=None):
        """
        Return the signature of a HTTP request.


        :param path_query: A string representation of the path and the query
            parameters of the URL endpoint referenced by the HTTP request.

        :param message_body: The body message passed to HTTP request using the
            method `POST` or `PUT`.


        :return: The signature of the HTTP request.
        """
        message = path_query if message_body is None else f'{path_query}{message_body}'

        signature = hmac.new(
            self.connection.consumer_secret.encode(),
            msg=message.encode(),
            digestmod=self.CRYPTOGRAPHIC_HASH_FUNCTION) \
            .hexdigest()

        return signature

    @staticmethod
    def __expand_url_bits(path, url_bits):
        """
        Find and replace names of URL bits by their corresponding value.


        :param path: Path of the Uniform Resource Identifier (URI) that names
            a resource of the platform the HTTP request requires to gain
            access to.

        :param url_bits: In-line arguments, also known as bits of URL.


        :return: A string representing the path of the URL, which URL bits
            have been expanded with their respective values.
        """
        for url_bit in re.findall(r'\([a-z0-9_]*\)', path):
            path = path.replace(url_bit, str(url_bits[url_bit[1:-1]]))

        return path

    def __get_headers(
            self,
            path_query,
            authentication_required=False,
            compatibility_required=False,
            files=None,
            headers=None,
            message_body=None,
            radioactive_tracer=False,
            signature_required=True):
        """
        Add additional HTTP headers to pass along the HTTP request.


        :param path_query: A string representation of the path and query
            parameters of the URL endpoint referenced by the HTTP request.

        :param authentication_required: Indicate whether the specified
            endpoint of the RESTful API server requires the user to be logged
            in.  This user session MUST correspond to a
            `PerseusAuthenticatedSession` or the function will raise an
            exception `ValueError`.  The functions adds a header
            'X-Authentication' corresponding to the identification of this
            user session.

        :param compatibility_required: Indicate whether the client application
            requires to use a specific version of the RESTful API server for
            this particular HTTP request.  The function adds a header
            'X-API-Version' with the value of the API version specified in the
             object `PerseusConnection` associated with this user session
             object.

        :param files: A list of files to be uploaded to the RESTful API server.

        :param headers: A dictionary of HTTP header to pass along the HTTP
            request.

        :param message_body: The body message passed to HTTP request using the
            method `POST` or `PUT`.

        :param radioactive_tracer: Indicate whether to generate an identifier

        :param radioactive_tracer: Indicate whether the HTTP request needs
            to integrate a "radioactive tracer", i.e., a nonce ("number used
            once"), a pseudo-random number, used to explore the mechanism of
            subsequent procedure calls by tracing the path that the HTTP
            request follows.  The function adds a header 'X-Radioactive-Tracer'.

        :param signature_required: Indicate whether to sign the HTTP request
            with the Consumer Secret defined in the object `PerseusConnection`
            associated with this user session object.  The functions adds a
            header 'X-API-Sig'.


        :return: A dictionary of headers.
        """
        if headers is None:
            headers = {}

        # headers['Content-Type'] = 'multipart/form-data' if files else 'application/json'
        if not files:
            headers['Content-Type'] = 'application/json'

        if radioactive_tracer or self.connection.radioactive_tracer:
            headers['X-Radioactive-Tracer'] = uuid.uuid1().hex

        if self.session_id:
            headers['X-Authentication'] = self.session_id.hex
        elif authentication_required:
            raise ValueError("A user session MUST be specified when authentication is required")

        headers['X-API-Key'] = self.connection.consumer_key
        if signature_required:
            headers['X-API-Sig'] = self.__calculate_signature(path_query, message_body=message_body)

        if compatibility_required:
            headers['X-API-Version'] = str(self.connection.api_version)

        return headers

    def __init__(self, connection, attributes=None):
        """
        Build an object `PerseusSession`.


        :param connection: An object `PerseusConnection` that references a
            Perseus compliant RESTful API server to connect to.

        :param attributes: An object that contains at least the following
            attributes:

            * `account_id` (required): Identification of the user as
              registered to the server platform.

            * `session_id` (required): Identification of the session of the
              user authenticated against the RESTful API server.

            * `expiration_time (required)`: Time when the login session of
              the authenticated user will expire.
        """
        if not isinstance(connection, PerseusConnection):
            raise ValueError('argument "connection" MUST be an object "PerseusConnection"')

        self.connection = connection
        self.session_id = attributes and attributes.session_id
        self.account_id = attributes and attributes.account_id
        self.expiration_time = attributes and attributes.expiration_time

    def __request__(
            self,
            http_method,
            path,
            url_bits=None,
            arguments=None,
            headers=None,
            message_body=None,
            files=None,
            authentication_required=False,
            signature_required=True,
            compatibility_required=False,
            radioactive_tracer=False,
            error_handler=None):
        """
        Send the specified HTTP request and return the response of the
        RESTful API server.


        :param http_method: An item of the enumeration `HttpMethod` that
            indicates the HTTP method of the request to send.

        :param path: The path of the Uniform Resource Identifier (URI) that
            names a resource of the platform the HTTP request requires to
            gain access to.

        :param url_bits: In-line arguments, also known as bits of URL.

        :param arguments: A dictionary of key/value pairs representing the
            arguments to be passed in the query string of the HTTP request.

        :param headers: A dictionary of key/value pairs representing the
            headers of the HTTP request.

        :param message_body: An object corresponding to the message body to be
            passed in the HTTP request using a HTTP method `POST` or `PUT`.

        :param files: A list of file-like objects to be uploaded.  The
            function will create a data stream "multipart/form-data" and add
            each file within a multipart MIME field named with a auto-
            generated UUID.

        :param authentication_required: Indicate whether the HTTP request must
            be sent on behalf of an authenticated user.  The HTTP request is
            added an additional header 'X-Authentication' with the access
            token (session ID) that impersonates this user.

        :param signature_required: Indicate whether the HTTP request must be
            signed.  The HTTP request is added two additional headers
            'X-API-Key' and 'X-API-Sig'.  The latter is calculated based on
            the content of the HTTP request and the API secret key.

        :param compatibility_required: Indicate whether the HTTP request must
            be strictly compatible with the version of the server platform's
            API.  The HTTP request is added the HTTP header field
            `X-API-Version` which value complies with a version number using
            a standard tuple of integers: `major.minor.patch`.

        :param radioactive_tracer: Indicate whether the HTTP request needs
            to integrate a "radioactive tracer", i.e., a nonce ("number used
            once"), a pseudo-random number, used to explore the mechanism of
            subsequent procedure calls by tracing the path that the HTTP
            request follows.  The HTTP request is added an additional header
            'X-Radioactive-Tracer'.

        :param error_handler: A Python function that will be passed any
            instance `urllib2.HTTPError` that represents an exceptional
            situation, i.e., a condition, basically an error, that the
            platform would face when performing the HTTP request.  This
            function is responsible for raising the appropriate Python
            exception.


        :return: A JSON object representing the response to this HTTP request.


        :raise SocketTimeoutException: If a timeout occurs on the socket
            during blocking operations like the connection attempt of the HTTP
            request.
        """
        # Define the complete URL of the endpoint referenced by the HTTP request.
        path = self.__expand_url_bits(path, url_bits)
        path_query = self.__build_path_query_string(path, arguments)
        url = f'{self.connection.scheme}://{self.connection.hostname}:{self.connection.port_number}{path_query}'
        stringified_message_body = None if http_method not in (HttpMethod.POST, HttpMethod.PUT) \
            else json.dumps(obj.stringify(message_body) if message_body else {})

        # Define the HTTP headers to pass along the HTTP request.
        headers = self.__get_headers(
            path_query,
            headers=headers,
            message_body=stringified_message_body,
            authentication_required=authentication_required,
            compatibility_required=compatibility_required,
            files=files,
            radioactive_tracer=radioactive_tracer,
            signature_required=signature_required)

        if self.connection.debug:
            logging.debug(self.__build_curl_command(http_method, path_query, message_body, headers))

        # Send the HTTP request.
        execution_begin_time = datetime.datetime.now()

        try:
            if files:
                files = dict([(uuid.uuid1().hex, f) for f in files])
                response = requests.put(url, files=files, headers=headers)
                data = response.text
            else:
                request = PerseusSession.Request(
                    url,
                    data=stringified_message_body and stringified_message_body.encode(),
                    headers=headers,
                    http_method=http_method)
                response = urllib.request.urlopen(request, timeout=self.connection.timeout)
                data = response.read()

            logging.debug(data)

        except urllib.request.HTTPError as error:
            if error_handler:
                data = error.read()
                logging.error(data)
                logging.exception(error)
                error_handler(error, None if (data is None or len(data) == 0) else json.loads(data))
            else:
                raise error

        except urllib.request.URLError as error:
            logging.exception(error)
            raise self.SocketTimeoutException() if isinstance(error.reason, socket.timeout) else error

        except socket.timeout as error:
            logging.exception(error)
            raise PerseusSession.SocketTimeoutException()

        # Log the latency time, i.e. the time delta between the instant the
        # client application sends the request to the RESTful API server and
        # the instant this client application receives the response.
        #
        # This duration includes the time the request and its response takes
        # to travel back and forth over the network between the client
        # application and the RESTful API server. The Latency corresponds to
        # Response Time plus the Network Time, also known as Time to First Byte
        # (TTFB) for the time it takes in milliseconds for a client application
        # to receive the first byte of the response from a server.
        execution_end_time = datetime.datetime.now()
        execution_duration = execution_end_time - execution_begin_time
        execution_duration_milliseconds = execution_duration.seconds * 1000 \
            + execution_duration.microseconds / 1000

        logging.debug(f"Query {http_method} {url} executed in {execution_duration_milliseconds}ms")

        # Return the JSON expression of the data that the RESTful API server
        # returned.
        return None if data is None or len(data) == 0 \
            else json.loads(data)

    @staticmethod
    def __stringify_query_parameter_value(value):
        """
        Stringify the value of a query parameter.


        If the argument `value` passed to this function is a list, a tuple, or
        a set, the function convert each item to a string, and concatenate all
        the strings separated with a comma. Otherwise, the function simply
        returns the string representation of the argument `value`:

        ```python
        >>> __stringify_query_parameter_value([1, 2, 3, 5, 7, 11])
        '1,2,3,4,5,7,11'
        >>> __stringify_query_parameter_value(uuid.UUID('2981c498-694a-11e3-ba63-00224d868489'))
        '2981c498-694a-11e3-ba63-00224d868489'
        ```

        :param value: An object.


        :return: The stringified value of this object.
        """
        # path_query_string = path if arguments is None \
        #     else '%s?%s' % \
        #          (path,
        #           urllib.parse.urlencode(dict([
        #              (key, ','.join([str(item).encode('utf-8') for item in value])
        #                  if type(value) in [list, set, tuple] else str(value).encode('utf-8'))
        #              for (key, value) in arguments.items()
        #              if value is not None])))

        return str(value) if type(value) not in (list, set, tuple) \
            else ','.join([str(element) for element in value])


class PerseusAnonymousSession(PerseusSession):
    """
    Anonymous user session connected to a RESTful API server.
    """
    def __init__(self, connection):
        """
        Build an object `PerseusAnonymousSession`.


        :param connection: An object `PerseusConnection` referencing a RESTful
            API server the user is connected to.
        """
        super().__init__(connection)


class PerseusAuthenticatedSession(PerseusSession):
    """
    Login session of a user connected to a RESTful API server.
    """
    def __init__(self, connection, attributes):
        """
        Build an object `PerseusAuthenticatedSession`.


        :param connection: An object `PerseusAuthenticatedSession` referencing
            a RESTful API server the user is connected to.

        :param attributes: An object that contains at least the following
            attributes:

            * `account_id` (required): Identification of the user as
              registered to the server platform.

            * `session_id` (required): Identification of the session of the
              user authenticated against the RESTful API server.

            * `expiration_time (required)`: Time when the login session of
              the authenticated user will expire.
        """
        if not isinstance(attributes, Object):
            raise ValueError('the argument "attributes" MUST inherit from `Object')
        super().__init__(connection, attributes=attributes)

    @staticmethod
    def from_payload(connection, payload):
        """
        Build an object `PerseusAuthenticatedSession` from the JSON
        serialization of a user login session.


        :param connection: An object `PerseusConnection` referencing the
            RESTful API server the user session belongs to.

        :param payload: A JSON expression containing the attribute of the user
            session.  This JSON expression MUST contain at least the following
            properties:

            * `account_id` (required): Identification of the user as
              registered to the server platform.

            * `session_id` (required): Identification of the session of the
              user authenticated against the RESTful API server.

            * `expiration_time (required)`: Time when the login session of
              the authenticated user will expire.


        :return: An object `PerseusAuthenticatedSession`.
        """
        if not payload:
            return None

        # Build the session's attributes from the JSON payload.
        attributes = cast.json_to_object(payload)

        # Convert the common attributes to their respective type.  The client
        # application will probably have to convert custom attributes.
        attributes.account_id = cast.string_to_uuid(payload['account_id'])
        attributes.expiration_time = cast.string_to_timestamp(payload['expiration_time'])
        attributes.locale = cast.string_to_locale(payload.get('locale'))
        attributes.picture_id = cast.string_to_uuid(payload.get('picture_id'))
        attributes.session_id = cast.string_to_uuid(payload['session_id'])

        # Return the authenticated session of the user.
        return PerseusAuthenticatedSession(connection, attributes)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from majormode.perseus.client.service.base_service import BaseService
from majormode.perseus.model.obj import Object


class NotificationService(BaseService):
    """
    A notification is a message that needs to be delivered to a set of
    recipients.  It informs the recipients about an event that occurs.
    """
    def __init__(self, connection):
        super(NotificationService, self).__init__(connection)

    def get_notifications(
            self,
            start_time=None, end_time=None,
            limit=BaseService.DEFAULT_RESULT_SET_SIZE,
            offset=0,
            include_read=False,
            mark_read=True):
        """
        Return a list of notifications that have been sent to the specified
        user's account, stored by ascending time of creation.

        @param start_time:
        @param offset: require to skip that many records before beginning to
               return records to the client.  Default value is ``0``.  If both
               ``offset`` and ``limit`` are specified, then ``offset`` records
               are skipped before starting to count the limit records that
               are returned.
        @param limit: constrain the number of records that are returned to the
               specified number.  Default value is ``20``. Maximum value is
               ``100``.
        @param include_read: indicate whether to include notifications that
               have already been read.  By default, notifications a user has
               read are not included.
        @param mark_read: indicate whether to mark read every notification
               that this functions returns.  By default, the function marks
               as read notifications that are returned.

        @return: a list of instances containing the following members:
                 * ``notification_id`` (required): identification of the
                   notification.
                 * ``notification_type`` (required): string representation of
                   the type of the notification, as defined by the client
                   application or the service that posted this notification to
                   the user.
                 * ``sender_id`` (optional): the identification of the account
                   of the user who initiates the notification.
                 * ``payload`` (optional): a dictionary containing additional
                   information specific to this particular notification.
                 * ``is_unread`` (required): indicate whether the notification
                   has been marked as read.
                 * ``creation_time`` (required): the time the notification was
                   originally sent.
                 * ``update_time`` (required): the time the notification was
                   originally sent, or the most recent time an attribute of
                   the notification, such as its read status, was updated.
       """
        return sorted(Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/notification',
                arguments={
                    'end_time': end_time,
                    'include_read': include_read,
                    'limit': limit,
                    'mark_read': mark_read,
                    'offset': offset,
                    'start_time': start_time
                },
                authentication_required=True,
                signature_required=True)),
            key=lambda notification: notification.creation_time)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.client.service.base_service import BaseService
from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.model.obj import Object


class AreaService(BaseService):
    def __init__(self, session):
        super(AreaService, self).__init__(session)

    def get_area_boundaries(self, area_id, sync_time=None):
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/area/(area_id)/boundaries',
                url_bits={
                    'area_id': area_id
                },
                arguments={
                    'sync_time': sync_time
                },
                authentication_required=False,
                signature_required=True))

    def get_areas_by_ip_address(
            self,
            ip_address,
            highest_area_level=None,
            locale=DEFAULT_LOCALE,
            lowest_area_level=None,
            minimal_area_surface=None):
        """
        Return the geographic areas that encompass the location determined
        from the specified IP address.


        @param ip_address: a dotted-decimal notation of an IPv4 address,
            consisting of four decimal numbers, each ranging from ``0`` to
            ``255``, separated by dots.

        @param lowest_area_level: the level of the smallest administrative
            area to start with.

            As a reminder, for clarity and convenience the standard neutral
            reference for the largest administrative subdivision of a country
            is called the "first-level administrative division" or "first
            administrative level". Next smaller is called "second-level
            administrative division" or "second administrative level", etc.

            Note: the smallest the area, the fastest the function returns
            result.

        @param highest_area_level: the level of the largest administrative
            area to finish with.

        @param locale: a ``Locale`` instance that specifies the locale of the
            area labels to be returned.

        @param minimal_area_surface: minimal surface in square meter of the
            geographical area that are returned.


        @return: a list of instances containing the following members:

            * ``area_id`` (required): identification of the geographic area.

            * ``area_label`` (required): an instance ``Label`` corresponding to the
              name of this area in the specified locale, or the closest locale if
              no name is defined for this particular locale, which is, at least,
              English by default.

            * ``area_level`` (required): administrative level of this geogaphic area.
              For clarity and convenience the standard neutral reference for the
              largest administrative subdivision of a country is called the "first-
              level administrative division" or "first administrative level".  Next
              smaller is called "second-level  administrative division" or "second
              administrative level".

            * ``area_type`` (optional): symbolic name of the type of this geographic
              area.

            * ``bounding_box`` (required): an instance ``BoundingBox`` representing
              the maximum extents of the geographical area.

            * ``centroid`` (required): an instance ``GeoPoint`` representing the
              centroid of this area.

            * ``parent_area_id`` (optional): identification of the parent geographic
              area or ``None`` if none defined.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/area',
                arguments={
                    'ip_address': '.'.join([str(byte) for byte in ip_address]),
                    'highest_area_level': highest_area_level,
                    'locale': locale,
                    'lowest_area_level': lowest_area_level,
                    'minimal_area_surface': minimal_area_surface
                },
                authentication_required=False,
                signature_required=True))

    def get_areas_by_keywords(
            self, keywords,
            highest_area_level=None,
            limit=None,
            locale=DEFAULT_LOCALE,
            lowest_area_level=None,
            minimal_area_surface=None,
            offset=0):
        """
        Return the geographic areas that match the specified keywords.


        @param keywords: a list of one or more keywords.

        @param highest_area_level: the level of the largest administrative
            area to finish with.

        @param limit: constrain the number of areas that are returned to the
            specified number.

        @param lowest_area_level: the level of the smallest administrative
            area to start with.

            As a reminder, for clarity and convenience the standard neutral
            reference for the largest administrative subdivision of a country
            is called the "first-level administrative division" or "first
            administrative level". Next smaller is called "second-level
            administrative division" or "second administrative level", etc.

            Note: the smallest the area, the fastest the function returns
            result.

        @param minimal_area_surface: minimal surface in square meter of the
            geographical area that are returned.

        @param offset: require to skip that many areas before beginning to
            return them.  If both ``limit`` and ``offset`` are specified, then
            ``offset`` areas are skipped before starting to count the
            `limit`` areas that are returned.  The default value is ``0``.

        @param locale: a ``Locale`` instance that specifies the locale of the
             area labels to be returned.


        @return: an instance containing the following members:

            * ``area_id``: identification of the geographic area.

            * ``area_type``: symbolic name of the type of the geographic area.

            * ``area_label``: label in the specified locale, or the closest locale
              if no label is defined for this particular locale, which is, at
              least, English by default.

            * ``parent_area_id``: identification of the parent geographic area or
              ``None`` if none defined.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/area',
                arguments={
                    'highest_area_level': highest_area_level,
                    'keywords': ','.join(keywords),
                    'limit': limit,
                    'locale': locale,
                    'lowest_area_level': lowest_area_level,
                    'minimal_area_surface': minimal_area_surface,
                    'offset': offset
                },
                authentication_required=False,
                signature_required=True))

    def get_areas_by_location(
            self,
            location,
            highest_area_level=None,
            locale=DEFAULT_LOCALE,
            lowest_area_level=None,
            minimal_area_surface=None):
        """
        Return the geographic areas that encompass the specified location.


        @param location: a ``GeoPoint`` instance that specifies a location.

        @param lowest_area_level: the level of the smallest administrative
            area to start with.

            As a reminder, for clarity and convenience the standard neutral
            reference for the largest administrative subdivision of a country
            is called the "first-level administrative division" or "first
            administrative level". Next smaller is called "second-level
            administrative division" or "second administrative level", etc.

            Note: the smallest the area, the fastest the function returns
            result.

        @param highest_area_level: the level of the largest administrative
            area to finish with.

        @param locale: a ``Locale`` instance that specifies the locale of the
            area labels to be returned.

        @param minimal_area_surface: minimal surface in square meter of the
            geographical area that are returned.


        @return: a list of instances containing the following members:

            * ``area_id`` (required): identification of the geographic area.

            * ``area_label`` (required): an instance ``Label`` corresponding to the
              name of this area in the specified locale, or the closest locale if
              no name is defined for this particular locale, which is, at least,
              English by default.

            * ``area_level`` (required): administrative level of this geogaphic area.
              For clarity and convenience the standard neutral reference for the
              largest administrative subdivision of a country is called the "first-
              level administrative division" or "first administrative level".  Next
              smaller is called "second-level  administrative division" or "second
              administrative level".

            * ``area_type`` (optional): symbolic name of the type of this geographic
              area.

            * ``bounding_box`` (required): an instance ``BoundingBox`` representing
              the maximum extents of the geographical area.

            * ``centroid`` (required): an instance ``GeoPoint`` representing the
              centroid of this area.

            * ``parent_area_id`` (optional): identification of the parent geographic
              area or ``None`` if none defined.
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/area',
                arguments={
                    'highest_area_level': highest_area_level,
                    'll': '%s,%s' % (location.longitude, location.latitude),
                    'locale': locale,
                    'lowest_area_level': lowest_area_level,
                    'minimal_area_surface': minimal_area_surface
                },
                authentication_required=False,
                signature_required=True))

    def get_areas_in_bounding_box(self, bounds, locale=DEFAULT_LOCALE):
        """

        @param bounds: a tuple of two instances ``GeoPoint`` that represent
            the north-east corner and the south-west corners of the rectangle
            area to search geographical areas in.

        @param locale: an instance ``Locale`` to return textual information of
            the places.


        @return:
        """
        return Object.from_json(
            self.send_request(
                http_method=self.HttpMethod.GET,
                path='/area',
                arguments={
                    'bounds': bounds and '%s,%s,%s,%s' % (
                        bounds[0].longitude, bounds[0].latitude,
                        bounds[1].longitude, bounds[1].latitude),
                    'locale': locale
                },
                authentication_required=False,
                signature_required=True))



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
# OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
# TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE SHALL NOT BE LIABLE FOR ANY
# LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
# OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import os

from majormode.perseus.client.api import PerseusConnection
from majormode.perseus.client.api import PerseusSession
from majormode.perseus.model.enum import Enum
from majormode.perseus.model.obj import Object
from majormode.perseus.utils import cast


class BaseService:
    """
    Base class of service.

    It provides logging facility to the inheriting service class.  The
    logger is identifier by the global constant ``LOGGER_NAME`` of
    the Python module ``settings``.

    It also provides facility to publish messages to an Advanced
    Message Queuing Protocol (AMQP) compliant broker, which connection
    properties are defined in the dictionary
    ``AMQP_CONNECTION_PROPERTIES`` defined in the Python module.
    ``settings``.
    """
    HttpMethod = Enum(
        'DELETE',
        'GET',
        'HEAD',
        'POST',
        'PUT'
    )

    class BaseServiceException(Exception):
        """
        Base class of exception that a service might raise to indicate some
        unexpected conditions.
        """
        def __init__(self, message=None, payload=None):
            super(BaseService.BaseServiceException, self).__init__(message)
            self.payload = payload

    class DeletedObjectException(BaseServiceException):
        """
        Signal that the specified object has been deleted and as such it
        cannot be used or referenced anymore.
        """
        pass

    class DisabledObjectException(BaseServiceException):
        """
        Signal that the specified object has been disabled and as such it
        cannot be used.
        """
        pass

    class ExpiredSessionException(BaseServiceException):
        """
        Signal that the account session of the user has expired.  The user
        needs to login again.
        """
        pass

    class InvalidOperationException(BaseServiceException):
        """
        Signal the operation requested is invalid, for instance one or several
        arguments provided within the request are not coherent together or the
        state of the object on which the operation is requested.
        """
        pass

    class IllegalAccessException(BaseServiceException):
        """
        Signal that the user on behalf whom a function is called is not
        authorized to perform the action requested.
        """
        pass

    class InvalidArgumentException(BaseServiceException):
        """
        Signal that an argument passed to a function is invalid, more likely
        of the wrong type.
        """
        pass

    class UndefinedObjectException(BaseServiceException):
        """
        Signal that the specified object doesn't exist and as such it cannot
        be referenced.
        """
        pass

    class MultipartFormData(object):
        """
        Represent data to be uploaded to the platform as in a multipart MIME
        data stream "multipart/form-data".
        """
        def __init__(self, field_name, file_name,
                file_path_name=None,
                data=None):
            """
            Build a new instance of this class.

            @param field_name: original field name in the form to send to the
                platform to upload a file.

            @param file_name: name of the file as defined in the HTTP request that
                the client application sent to the platform.

            @param file_path_name: complete file path name of the file to upload.

            # @param content_type: specify the nature of the data in the content of
            #     the file that has been uploaded to the platform entity, by giving
            #     type and subtype identifiers, and by providing auxiliary
            #     information that may be required for certain types.

            @param data: content of the file that has been uploaded.
            """
            if data is None and file_path_name is None:
                raise self.InvalidArgumentException("Either data MUST be not null, either a file path name MUST be specified")

            self.field_name = field_name
            self.file_name = file_name
            self.file_path_name = file_path_name
            self.data = data

        @staticmethod
        def from_file(file_path_name, field_name=None):
            file_name = os.path.basename(file_path_name)
            return BaseService.MultipartFormData(
                    field_name or os.path.splitext(file_name)[0],
                    file_name,
                    file_path_name=file_path_name)

        def read_data(self):
            if self.data is not None:
                return self.data


    BASE_SERVICE_EXCEPTIONS = {
        'DeletedObjectException': DeletedObjectException,
        'DisabledObjectException': DisabledObjectException,
        'ExpiredSessionException': ExpiredSessionException,
        'InvalidOperationException': InvalidOperationException,
        'IllegalAccessException': IllegalAccessException,
        'InvalidArgumentException': InvalidArgumentException,
        'UndefinedObjectException': UndefinedObjectException
    }

#    DEFAULT_LIMIT = 20
#    MAXIMUM_LIMIT = 100

    # Default number of items that a remote method would return.
    DEFAULT_RESULT_SET_SIZE = 20

    # Maximum number of items that a remote method would return.
    MAXIMUM_RESULT_SET_SIZE = 100

    def __init__(self, session_or_connection):
        """
        Build an instance ``BaseService``.

        @param session_or_connection: an instance of a class that inherits
            from ``PerseusSession``, such as ``PerseusAnonymousSession`` or
            ``PerseusAuthenticatedSession``.  This constructor also accepts an
            instance of a class that inherits from ``PerseusConnection``; in
            that case, this constructor builds an anonymous session.
        """
        if session_or_connection is None:
            raise ValueError('A session or a connection properties object MUST be passed')

        self._service_name = self.__class__.__name__

        if isinstance(session_or_connection, PerseusSession):
            self.session = session_or_connection
        elif isinstance(session_or_connection, PerseusConnection):
            self.session = session_or_connection.build_session()
        else:
            raise ValueError("The class of the instance session MUST inherits from 'PerseusSession' or 'PerseusConnection'")

    @staticmethod
    def _declare_custom_exceptions(exceptions):
        BaseService.BASE_SERVICE_EXCEPTIONS.update(exceptions)

    def __default_error_handler__(self, error, payload):
        """
        Default handler of error the server platform raises when a HTTP
        request fails.

        @param error: an instance ``urllib2.HTTPError``.
        """
        try:
            exception_name = payload['error']
            exception_class = BaseService.BASE_SERVICE_EXCEPTIONS[exception_name]

            raise exception_class(
                    message=payload.get('message'),
                    payload=payload.get('payload'))

        except KeyError:
            raise Exception('Undeclared platform service exception "%s"' % payload['error'])

    @staticmethod
    def decode_object(obj, cast_operators):
        def __decode_value__(value, cast_operator=None):
            return value if value is None or cast_operator is None else \
                cast.string_to_enum(value, cast_operator) if isinstance(cast_operator, enum.Enum) \
                else cast_operator(value)

        return Object(**dict([ (name, __decode_value__(value, cast_operators.get(name)))
                 for (name, value) in obj.__dict__.iteritems() ]))

    def flush_session(self):
        self.session = self.session.connection.build_session()

    def get_session(self):
        return self.session

    def get_service_name(self):
        return self._service_name

    def send_request(self,
            http_method=HttpMethod.GET,
            headers=None,
            path=None,
            url_bits=None,
            arguments=None,
            message_body=None,
            files=None,
            authentication_required=False,
            signature_required=True,
            compatibility_required=False,
            radioactive_tracer=False):
        return self.session.__request__(
                http_method=http_method,
                headers=headers,
                path=path,
                url_bits=url_bits,
                arguments=arguments,
                files=files,
                message_body=message_body,
                authentication_required=authentication_required,
                signature_required=signature_required,
                compatibility_required=compatibility_required,
                radioactive_tracer=radioactive_tracer,
                error_handler=self.__default_error_handler__)

    def set_session(self, session):
        """
        Redefine the session used by the instance of this inherited service,
        for instance, when the user signs-in or when he signs-out.

        @param session: an instance of a class that inherits from
            ``PerseusSession``, such as ``PerseusAnonymousSession`` or
            ``PerseusAuthenticatedSession``.
        """
        if session and not isinstance(session, PerseusSession):
            raise Exception('The class of the instance session MUST inherits from ``PerseusSession``')

        self.session = session
